/* ***** BEGIN LICENSE BLOCK *****
 *   Version: MPL 1.1/GPL 2.0/LGPL 2.1
 *
 * The contents of this file are subject to the Mozilla Public License Version
 * 1.1 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/
 * 
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is ThumbsDown.
 *
 * The Initial Developer of the Original Code is
 * Atte Kemppilä.
 * Portions created by the Initial Developer are Copyright (C) 2008
 * the Initial Developer. All Rights Reserved.
 *
 * Contributor(s):
 *
 * Alternatively, the contents of this file may be used under the terms of
 * either the GNU General Public License Version 2 or later (the "GPL"), or
 * the GNU Lesser General Public License Version 2.1 or later (the "LGPL"),
 * in which case the provisions of the GPL or the LGPL are applicable instead
 * of those above. If you wish to allow use of your version of this file only
 * under the terms of either the GPL or the LGPL, and not to allow others to
 * use your version of this file under the terms of the MPL, indicate your
 * decision by deleting the provisions above and replace them with the notice
 * and other provisions required by the GPL or the LGPL. If you do not delete
 * the provisions above, a recipient may use your version of this file under
 * the terms of any one of the MPL, the GPL or the LGPL.
 * 
 * ***** END LICENSE BLOCK ***** */


const Cc = Components.classes;
const Ci = Components.interfaces;
const Cr = Components.results;

const atomSvc = Cc["@mozilla.org/atom-service;1"].getService(Ci.nsIAtomService);
const ioSvc = Cc["@mozilla.org/network/io-service;1"].getService(Ci.nsIIOService);
const scriptMgr = Cc["@thumbsdown.mozdev.org/script-manager;1"].getService(Ci.nsISupports).wrappedJSObject;

const MODE_CREATE = 1;
const MODE_MODIFY = 2;

const TYPE_DIRECT = 0;
const TYPE_REGEXP = 1;
const TYPE_CUSTOM = 2;

var wizard = null;
var pages = {};
var page = null;

var isAdvancing = false;


function init()
{
    wizard = $("thumbsdown_scriptWizard");
    stringBundle = $("stringbundle");
    model.init();
}

function onUnload()
{
    for each (thumbnail in model.thumbnails)
    {
        if (thumbnail.targetRequest.timeoutId)
        {
            window.clearTimeout(thumbnail.targetRequest.timeoutId);
            delete thumbnail.targetRequest.timeoutId;
        }
        thumbnail.targetRequest.abort();
    }
}

function onPageShow(e)
{
    var pageId = e.target.pageid;
    page = pages[pageId];
    if (page.onPageShow)
        page.onPageShow();
    updateWizardButtons();
}

function onPageAdvanced()
{
    if (isAdvancing)
        isAdvancing = false;
    else if (page.onPageAdvanced)
        return page.onPageAdvanced();
    return true;
}

function onPageRewound()
{
    if (page.onPageRewound)
        return page.onPageRewound();
    return true;
}

function advance(pageId)
{
    isAdvancing = true;
    wizard.advance(pageId);
    return false;
}

function updateWizardButtons()
{
    wizard.canAdvance = (page.canAdvance ? page.canAdvance() : true);
    wizard.canRewind  = (page.canRewind  ? page.canRewind()  : true);
}


var model =
{
    get thumbnails() { return this._thumbnails; },
    set thumbnails(value)
    { 
        this._thumbnails = value;
        this.selected = null;
    },
    
    get selected() { return this._selected; },
    set selected(value)
    { 
        this._selected = value;
        this.match = null;
        this.linkUrlTokens = null;
        this.mode = null;
    },
    
    get mode() { return this._mode; },
    set mode(value)
    {
        if (value == this._mode)
            return;
        
        this._mode = value;
        this.scriptId = null;
    },
    
    get isCreate() { return (this.mode == MODE_CREATE); },
    get isModify() { return (this.mode == MODE_MODIFY); },
    
    get scriptId() { return this._scriptId; },
    set scriptId(value)
    {
        if (value == this._scriptId)
            return;
        
        this._scriptId = value;
        this.name0 = null;
        this.matcher0 = null;
        this.type = null;
    },
    
    get script() { return scriptMgr.getScript(this.scriptId); },
    
    get name0() { return this._name0; },
    set name0(value)
    {
        this._name0 = value;
        this.name = null;
    },
    
    get matcher0() { return this._matcher0; },
    set matcher0(value)
    {
        this._matcher0 = value;
        this.matcher = null;
    },
    
    get isDirect() { return (this.type == TYPE_DIRECT); },
    get isRegExp() { return (this.type == TYPE_REGEXP); },
    get isCustom() { return (this.type == TYPE_CUSTOM); },
    
    isInitialized: false,
    
    init: function()
    {
        if (this.isInitialized)
            return;
        this.isInitialized = true;
        
        this.thumbnails = [];
        var thumbnails = this.thumbnails;
        var doc = window.arguments[0];
        
        if (! doc.images)
            return;
        
        for (var i = 0; i < doc.images.length; i++)
        {
            var image = doc.images[i];
            var anchor = image.parentNode;
            if (anchor && anchor.href && (anchor instanceof HTMLAnchorElement))
            {
                var thumbnail =
                {
                    image: image,
                    anchor: anchor,
                    thumbnailUrl: ioSvc.newURI(image.src, null, null).QueryInterface(Ci.nsIURL),
                    linkUrl: ioSvc.newURI(anchor.href, null, null).QueryInterface(Ci.nsIURL),
                    index: thumbnails.length
                }
                thumbnail.targetRequest = this._startTargetRequest(thumbnail);
                thumbnails.push(thumbnail);
            }
        }
        
        // TODO: keep name list in scriptMgr
        this.names = scriptMgr.sqlExec.fetchAll("getNames");
    },
    
    _startTargetRequest: function(thumbnail)
    {
        function onThumbnailTargetUpdate(thumbnail)
        {
            if (page && page.onThumbnailTargetUpdate)
                page.onThumbnailTargetUpdate(thumbnail);
        }
        
        function onError(e)
        {
            if (thumbnail.target)
                return;
            
            var request = e.target;
            
            if (request.timeoutId)
            {
                window.clearTimeout(request.timeoutId);
                delete request.timeoutId;
            }
            
            thumbnail.target = {success: false, type: "error"};
            
            /*
             * If request fails and wizard page has been changed already, you get:
             * 0x80040111 (NS_ERROR_NOT_AVAILABLE) [nsIXMLHttpRequest.status]
             * https://bugzilla.mozilla.org/show_bug.cgi?id=301705
             */
            try { thumbnail.target.status = request.status; }
            catch (e if (e.result == Cr.NS_ERROR_NOT_AVAILABLE)) { /* ignore */ }
            
            onThumbnailTargetUpdate(thumbnail);
        }
        
        function onLoad(e)
        {
            if (thumbnail.target)
                return;
            
            var request = e.target;
            
            if (request.status != 200)
            {
                onError(e);
                return;
            }
            
            if (request.timeoutId)
            {
                window.clearTimeout(request.timeoutId);
                delete request.timeoutId;
            }
            
            thumbnail.target = {success: true, type: "unknown", status: request.status};
            var contentType = thumbnail.contentType = request.getResponseHeader("Content-Type");
            if (contentType)
            {
                var match =  /^([-\w]+\/[-\w]+)/.exec(contentType);
                if (match)
                {
                    var mimeType = thumbnail.target.mimeType = match[1];
                    if (/^image\/\w+/.test(mimeType))
                        thumbnail.target.type = "image";
                    else if (mimeType == "text/html")
                        thumbnail.target.type = "page";
                }
            }
            onThumbnailTargetUpdate(thumbnail);
        }
        
        function onTimeout(request)
        {
            request.abort();
            delete request.timeoutId;
            
            if (thumbnail.target)
                return;
            
            thumbnail.target = {success: false, type: "timeout"};
            onThumbnailTargetUpdate(thumbnail);
        }
        
        var request = new XMLHttpRequest();
        request.open("HEAD", thumbnail.anchor.href, true);
        request.onload = onLoad;
        request.onerror = onError;
        request.send(null);
        
        var timeout = nsPreferences.getIntPref("extensions.thumbsdown.scriptwizard.requestTimeout");
        request.timeoutId = window.setTimeout(function() { onTimeout(request); }, timeout * 1000);
        
        return request;
    }
}


pages.thumbnails = 
{
    isInitialized: false,
    tree: null,
    thumbnailIndex: null,
    
    get targetCount()
    {
        if (! this._targetCount)
        {
            this._targetCount = {image: 0, page: 0, unknown: 0, error: 0};
            for each (thumbnail in model.selected)
                if (thumbnail.target)
                    this._targetCount[thumbnail.target.type]++;
        }
        return this._targetCount;
    },
    
    onPageShow: function()
    {
        if (! this.isInitialized)
        {
            init();
            
            this.isInitialized = true;
            this.tree = $("tree_thumbnails");
            this.tree.view = new ThumbnailTreeView(model.thumbnails);
            
            this._updateSelected();
        }
    },
    
    onPageAdvanced: function()
    {
        if (model.match)
            return advance("script");
        
        if (! model.linkUrlTokens)
        {
            // try to find an existing script that matches all the select thumbnails
            var matchesArray = [];
            for each (thumbnail in model.selected)
            {
                var matches = scriptMgr.lookupAll(thumbnail.anchor.href, thumbnail.image.src);
                if (matches.length == 0)
                    break;
                matchesArray.push(matches);
            }
            
            if (matchesArray.length == model.selected.length)
            {
                for (var i = 0; i < matchesArray[0].length; i++)
                {
                    var script = matchesArray[0][i];
                    var isSameScript = function(s) { return (s.id == script.id); };
                    for (var k = 1; k < matchesArray.length; k++)
                    {
                        var matches = matchesArray[k];
                        if (! matches.some(isSameScript))
                            break;
                    }
                    
                    if (k == matchesArray.length)
                    {
                        // matching script has been found
                        model.match = script;
                        return advance("script");
                    }
                }
            }
            
            var linkUrls = [thumbnail.linkUrl for each (thumbnail in model.selected)];
            var urlTokens = new UrlTokens(linkUrls);
            
            try
            {
                if (! urlTokens.schemes.equal)
                    throw "scheme.notEqual";
                
                if (! urlTokens.hostTokens.slice(1).every(function(sa) { return sa.equal; }))
                    throw "host.notEqual";
                
                if (urlTokens.pathTokens[urlTokens.pathTokens.length - 1].hasNull)
                    throw "path.notEqualCount";
                
                model.linkUrlTokens = urlTokens;
                model.type = TYPE_DIRECT;
            }
            catch (e if (typeof(e) == "string"))
            {
                model.error = e;
                return advance("error");
            }
        }
        
        return true;
    },
    
    onThumbnailTreeSelect: function()
    {
        this._updateSelected();
        updateWizardButtons();
    },
    
    onThumbnailTreeKeyUp: function()
    {
        this._setThumbnail(this.tree.currentIndex);
    },
    
    // XXX: mouse scroll event doesn't work properly in ff2: https://bugzilla.mozilla.org/show_bug.cgi?id=352179
    
    onThumbnailTreeMouseMove: function(e)
    {
        var row = {};
        var col = {};
        var childElement = {};
        this.tree.treeBoxObject.getCellAt(e.clientX, e.clientY, row, col, childElement);
        var row = row.value;
        if (row >= 0)
            this._setThumbnail(row);
    },
    
    onThumbnailTreeMouseOut: function(e)
    {
        this._setThumbnail(null);
    },
    
    onThumbnailTargetUpdate: function(thumbnail)
    {
        var row = thumbnail.index;
        var col = this.tree.columns["tcol_thumbnails_target"];
        this.tree.boxObject.invalidateCell(row, col);
        this._targetCount = null;
        updateWizardButtons();
    },
    
    canAdvance: function()
    {
        if (model.selected.length < 2)
            return false;
        var count = this.targetCount;
        if ((count.unknown + count.error + count.page) > 0)
            return false;
        return true;
    },
    
    _updateSelected: function()
    {
        model.selected = [model.thumbnails[row] for each (row in getTreeSelectionIterator(this.tree.view.selection))];
        this._targetCount = null;
        $("lbl_thumbnails_treeStatus").value = _("count", model.selected.length, model.thumbnails.length);
    },
    
    _setThumbnail: function(index)
    {
        if (index == this.thumbnailIndex)
            return;
        
        this.thumbnailIndex = index;
        
        var image = $("image_thumbnail");
        if (index != null)
        {
            var thumbnail = model.thumbnails[index];
            
            var boxComputedStyle = document.defaultView.getComputedStyle($("box_thumbnail"), null);
            var boxWidth = boxComputedStyle.width;
            var boxHeight = boxComputedStyle.height;
            boxWidth = Number(boxWidth.substr(0, boxWidth.length - 2));     // "- 2": strip 'px'
            boxHeight = Number(boxHeight.substr(0, boxHeight.length - 2));
            var thumbnailWidth = thumbnail.image.naturalWidth;
            var thumbnailHeight = thumbnail.image.naturalHeight;
            var factor = Math.min(1, Math.min(boxWidth / thumbnailWidth, boxHeight / thumbnailHeight));
            
            image.src = thumbnail.image.src;
            image.style.width = (factor * thumbnailWidth) + "px";
            image.style.height = (factor * thumbnailHeight) + "px";
            $("lbl_thumbnailSize").value = _("thumbnailSize", thumbnailWidth, thumbnailHeight);
        }
        else
        {
            image.src = "";
            $("lbl_thumbnailSize").value = "";
        }
    }
}


pages.mode =
{
    tree: null,
    
    get mode()      { return $("rgroup_mode").value; },
    set mode(value) { $("rgroup_mode").value = value; },
    
    get isCreate() { return (this.mode == MODE_CREATE); },
    get isModify() { return (this.mode == MODE_MODIFY); },
    
    get scriptId()      { return Number($("mlist_mode_scripts").value); },
    set scriptId(value) { $("mlist_mode_scripts").value = value; },
    
    get script() { return (this.scriptId ? scriptMgr.getScript(this.scriptId) : null); },
    
    onPageShow: function()
    {
        if (! model.mode)
        {
            this.tree = $("tree_mode");
            this.tree.view = new ThumbnailTreeView(model.selected);
            
            var popup = $("mpopup_mode_scripts");
            while (popup.hasChildNodes())
                popup.removeChild(popup.firstChild);
            
            popup.appendChild(createElement("menuitem", {label: "", value: 0}));
            $("mlist_mode_scripts").selectedIndex = 0;
            
            for each (script in model.names)
                popup.appendChild(createElement("menuitem", {label: script.name, value: script.id}));
            
            this._updateScript();
        }
    },
    
    onPageAdvanced: function()
    {
        model.mode = this.mode;
        model.scriptId = (model.isModify ? this.scriptId : null);
        // XXX: what if new type != old type
        if (model.isModify)
            model.type = model.script.type;
        return true;
    },
    
    onModeSelect: function()
    {
        this._updateScript();
        updateWizardButtons();
    },
    
    onScriptSelect: function()
    {
        this._updateScript();
        updateWizardButtons();
    },
    
    canAdvance: function()
    {
        if (this.isCreate)
            return true;
        if (this.isModify && this.script && (this.script.type == TYPE_DIRECT))
            return true;
        return false;
    },
    
    _updateScript: function()
    {
        $("mlist_mode_scripts").setAttribute("disabled", (! this.isModify));
        
        if (this.isModify && this.script)
        {
            $("tbox_mode_scriptMatcher").value = this.script.regexp;
            $("tbox_mode_scriptType").value = _("type." + this.script.type);
        }
        else
        {
            $("tbox_mode_scriptMatcher").value = " ";
            $("tbox_mode_scriptType").value = " ";
        }
    }
}


pages.name =
{
    tree: null,
    
    get name()      { return $("tbox_name").value; },
    set name(value) { $("tbox_name").value = value; },
    
    onPageShow: function()
    {
        if (! model.name0)
        {
            this.tree = $("tree_name");
            this.tree.view = new ThumbnailTreeView(model.selected);
            
            model.name0 = (model.isModify ? model.script.name : this._resolveName());
            this.name = model.name0;
            
            this._updateButtons();
        }
    },
    
    onPageAdvanced: function()
    {
        model.name = this.name;
        return true;
    },
    
    onNameInput: function()
    {
        this._updateButtons();
        updateWizardButtons();
    },
    
    resetName: function()
    {
        this.name = model.name0;
        this._updateButtons();
    },
    
    canAdvance: function()
    {
        return (this.name.length > 0);
    },
    
    _updateButtons: function()
    {
        $("btn_name_reset").disabled = (this.name == model.name0);
    },
    
    _resolveName: function()
    {
        var urlTokens = model.linkUrlTokens;
        var hostToken = (urlTokens.hostTokens[0].equal ? urlTokens.hostTokens[0].strings[0] : null);
        var index = ((hostToken && (! /^www/.test(hostToken))) ? 0 : 1);
        var name = [sa.strings[0] for each (sa in urlTokens.hostTokens.slice(index))].join(".");
        return name;
    }
}


pages.matcher =
{
    tree: null,
    treeView: null,
    
    get matcher() { return $("tbox_matcher").value; },
    set matcher(value)
    {
        $("tbox_matcher").value = value;
        this._matcherRegExp = null;
    },
    
    get matcherRegExp()
    {
        if (! this._matcherRegExp)
            try { this._matcherRegExp = new RegExp("^" + this.matcher + "$"); }
            catch (e) { /* ignore */ }
        return this._matcherRegExp;
    },
    
    onPageShow: function()
    {
        if (! model.matcher0)
        {
            this.tree = $("tree_matcher");
            this.treeView = new ThumbnailTreeView(model.selected);
            this.tree.view = this.treeView;
            
            model.matcher0 = (model.isModify ? model.script.regexp : this._resolveMatcher());
            this.matcher = model.matcher0;
            
            this.onMatcherInput();
        }
    },
    
    onPageAdvanced: function()
    {
        model.matcher = this.matcher;
        return advance("script");
    },
    
    onMatcherInput: function()
    {
        this._matcherRegExp = null;
        if (this.matcherRegExp)
        {
            var matches = [this.matcherRegExp.test(thumbnail.anchor.href) for each (thumbnail in model.selected)];
            this.matchCount = count(matches);
        }
        else
        {
            var matches = [false for each (thumbnail in model.selected)];
            this.matchCount = 0;
        }
        this.treeView.rowStates = matches;
        this._update();
    },
    
    resetMatcher: function()
    {
        this.matcher = model.matcher0;
        this.onMatcherInput();
    },
    
    canAdvance: function()
    {
        return (this.matchCount == model.selected.length);
    },
    
    _update: function()
    {
        $("btn_matcher_reset").disabled = (this.matcher == model.matcher0);
        $("lbl_matcher_treeStatus").value = _("count", this.matchCount, model.selected.length);
        updateWizardButtons();
    },
    
    _resolveMatcher: function()
    {
        var urlTokens = model.linkUrlTokens;
        var sentinels;
        var tokens;
        
        // scheme
        var scheme = resolveRegExp(urlTokens.schemes);
        
        // host
        sentinels = arrayWithValue(urlTokens.hostTokens.length, ".");
        sentinels[sentinels.length - 1] = "/";
        tokens = [];
        for (var i = 0; i < urlTokens.hostTokens.length; i++)
            tokens.push(resolveRegExp(urlTokens.hostTokens[i], sentinels[i]));
        var host = tokens.join(escapeRegExp("."));
        
        // path
        sentinels = arrayWithValue(urlTokens.pathTokens.length, "/");
        sentinels[sentinels.length - 1] = null;
        tokens = [];
        for (var i = 0; i < urlTokens.pathTokens.length; i++)
            tokens.push(resolveRegExp(urlTokens.pathTokens[i], sentinels[i]));
        var path = tokens.join(escapeRegExp("/"));
        
        // query
        sentinels = arrayWithValue(urlTokens.queryTokens.length, "&");
        sentinels[sentinels.length - 1] = null;
        tokens = [];
        for (var i = 0; i < urlTokens.queryTokens.length; i++)
            tokens.push(resolveRegExp(urlTokens.queryTokens[i], sentinels[i]));
        var query = tokens.join(escapeRegExp("&"));
        
        return scheme + "://" + host + path + query;
    }
}

pages.error =
{
    tree: null,
    
    onPageShow: function()
    {
        $("lbl_error_message").textContent = _("error." + model.error);
        
        this.tree = $("tree_error");
        this.tree.view = new ThumbnailTreeView(model.selected);
    },
    
    onPageRewound: function()
    {
        model.error = null;
        return true;
    },
    
    canAdvance: function()
    {
        return false;
    }
}


pages.script =
{
    onPageShow: function()
    {
        var script = model;
        var msg = _("finish.save");
        
        if (model.match)
        {
            script = model.match;
            msg = _("finish.match");
        }
            
        $("lbl_script_msg").textContent = msg;
        $("tbox_script_name").value = script.name;
        $("tbox_script_matcher").value = script.matcher;
        $("tbox_script_type").value = _("type." + script.type);
    }
}


function onWizardFinish()
{
    if (model.match)
        return true;
    
    var script = (model.isCreate ? {} : model.script);
    script.name = model.name;
    script.regexp = model.matcher;
    script.type = model.type;
    if (model.isCreate)
        scriptMgr.addScript(script);
    else
        scriptMgr.setScript(script);
    
    return true;
}


function UrlTokens(urls)
{
    this.urls = urls;
    
    this.schemes = new StringArray([url.scheme for each (url in urls)]);
    this.hosts   = new StringArray([url.host   for each (url in urls)]);
    this.paths   = new StringArray([url.path   for each (url in urls)]);
    this.queries = new StringArray([url.query  for each (url in urls)]);
    
    this.hostTokens = this.hosts.tokenize(".");
    this.pathTokens = this.paths.tokenize("/");
    this.queryTokens = this.queries.tokenize("&");
}


function resolveRegExp(sa, sentinel)
{
    if (sa.equal)
        return escapeRegExp(sa.strings[0]);
    
    function resolve(sa, sentinel)
    {
        var group = null;
        var chars = "";
        var pattern = null;
        var multiplier = ((sa.length.min == 0) ? "*" : "+");
        
        var i = 0, k = 0;
        do
        {
            var c = sa.strings[i][k];
            
            if ((! group) && /\d/.test(c))
                group = "\\d";
            else if ((group != "\\w") && /\w/.test(c))
                group = "\\w";
            else if (chars.length < 1)
                chars = ((c == "-") ? c + chars : chars + c);
            else if (! pattern)
                pattern = (sentinel ? "[^" + sentinel + "]" : ".");
            else
                throw new Error("unreachable code");
            
            var p = (pattern ? pattern : (chars ? "[" + escapeRegExp(chars) + group + "]" : group)) + multiplier;
            var re = new RegExp(p);
            
            for (i = 0; i < sa.strings.length; i++)
            {
                var result = re.exec(sa.strings[i]);
                if (result[0] != sa.strings[i])
                {
                    k = ((result.index = 0) ? sa.strings[i][result[0].length] : 0);
                    break;
                }
            }
        }
        while (i < sa.strings.length);
        
        return p;
    }
    
    var prefix = escapeRegExp(sa.prefix);
    var middle = resolve(sa.middle(), sentinel);
    var suffix = escapeRegExp(sa.suffix);
    
    return prefix + middle + suffix;
}


function ThumbnailTreeView(thumbnails)
{
    this.thumbnails = thumbnails;
}

ThumbnailTreeView.prototype =
{
    get rowStates() { return this._rowStates; },
    set rowStates(value)
    {
        this.treeBoxObject.beginUpdateBatch();
        var firstFail = null;
        if (this._rowStates !== undefined)
            for (var row = 0; row < value.length; row++)
            {
                if (value[row] != this._rowStates[row])
                    this.treeBoxObject.invalidateRow(row);
                if ((firstFail == null) && (! value[row]))
                    firstFail = row;
            }
        if (firstFail != null)
            this.treeBoxObject.ensureRowIsVisible(firstFail);
        this._rowStates = value;
        this.treeBoxObject.endUpdateBatch();
    },
    
    get rowCount() { return this.thumbnails.length; },
    
    setTree: function(treeBoxObject)
    {
        this.treeBoxObject = treeBoxObject;
    },
    
    getCellText: function(row, col)
    {
        var thumbnail = this.thumbnails[row];
        switch (col.id)
        {
            case "tcol_thumbnails_link":
            case "tcol_mode_link":
            case "tcol_name_link":
            case "tcol_matcher_link":
            case "tcol_error_link":
                return thumbnail.anchor.href;
            
            case "tcol_thumbnails_target":
                var target = thumbnail.target;
                if (! target)
                    return "";
                switch (target.type)
                {
                    case "image":   return _("target.image");
                    case "page":    return _("target.page");
                    case "unknown": return (target.contentType ? _("target.unknown", (target.mimeType ? target.mimeType : target.contentType)) : _("target.noContentType"));
                    case "error":   return _("target.error", (target.status ? target.status : -1));
                    case "timeout": return _("target.timeout");
                }
                throw new Error("unknown target: " + target.type);
        }
        throw new Error("unknown column: " + col.id);
    },
    
    getRowProperties: function(row, props)
    {
        if (this.rowStates !== undefined)
            props.AppendElement(this._getAtom(this.rowStates[row] ? "pass" : "fail"));
    },
    
    isContainer: function(row) { return false; },
    isSeparator: function(row) { return false; },
    isSorted: function() { return false; },
    getLevel: function(row) { return 0; },
    getImageSrc: function(row, col) { return null; },
    getCellProperties: function(row, col, props) {},
    getColumnProperties: function(col, props) {},
    
    // http://www.xulplanet.com/tutorials/xultu/treestyle.html
    _atoms: {},
	_getAtom: function(a)
    {
		if (! (a in this._atoms))
			this._atoms[a] = atomSvc.getAtom(a);
		return this._atoms[a];
	}
}
